/* Copyright 2006 Nicholas Bishop
 * 
 * This file is part of SharpConstruct.
 *
 * SharpConstruct is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * SharpConstruct is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SharpConstruct; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

#ifndef BRUSHMODE_HH
#define BRUSHMODE_HH

#include "Color.hh"
#include "Math.hh"
#include "MeshHistory.hh"
#include "Point.hh"
#include <gdkmm/pixbuf.h>
#include <gtkmm/treepath.h>

namespace SharpConstruct
{
	namespace Brush
	{
		class SupportedProperties
		{
		public:
			SupportedProperties( const bool, const bool,
					     const bool, const bool,
					     const bool, const bool,
					     const bool, const bool,
					     const bool );

			const bool Strength, Intensity, Layer, Dot, Mask,
				   Add, Sub, Colorize, Pigment;
		};

		enum DirectionData{ OFF, IN, OUT };

		class EditData
		{
		public:
			unsigned SymmetryIndex;
			Normal3D Up, Left;
			Point3D Center;
		};

		class SymmetryData
		{
		public:
			SymmetryData();
			SymmetryData( const bool, const bool, const bool );
			
			bool X() const;
			void SetX( const bool );
			bool Y() const;
			void SetY( const bool );
			bool Z() const;
			void SetZ( const bool );
			
		private:
			bool x_, y_, z_;
		};

		class BrushBase
		{
		public:
			BrushBase();
			virtual ~BrushBase();

			virtual SupportedProperties Properties() const = 0;
			virtual bool SingleSelect() const;

			virtual DirectionData Direction() const;
			virtual void SetDirection( const DirectionData );

			virtual bool Colorize() const;
			virtual void SetColorize( const bool );

			virtual bool Mask() const;
			virtual void SetMask( const bool );

			virtual bool AffectsVertexData() const;
			virtual bool AffectsColorData() const;
			virtual bool AffectsPolygonData() const;

			virtual unsigned Radius() const;
			unsigned BasicRadius() const;
			virtual void SetRadius( const unsigned );

			virtual float Strength() const;
			float BasicStrength() const;
			virtual void SetStrength( const float );

			virtual float Intensity() const;
			float BasicIntensity() const;
			virtual void SetIntensity( const float );

			virtual bool Flip() const;
			virtual void SetFlip( const bool );

			virtual float Pressure() const;
			virtual void SetPressure( const float );

			virtual bool RadiusPressure() const;
			virtual void SetRadiusPressure( const bool );

			virtual bool StrengthPressure() const;
			virtual void SetStrengthPressure( const bool );

			virtual bool IntensityPressure() const;
			virtual void SetIntensityPressure( const bool );

			virtual Color Pigment() const;
			virtual void SetPigment( const Color );

			static SymmetryData Symmetry();
			static void SetSymmetry( const SymmetryData );

			virtual bool Dot() const;
			virtual void SetDot( const bool );
			
			virtual bool Layer() const;
			virtual void SetLayer( const bool );

			Gtk::TreePath ShapePath() const;
			virtual void SetShape( Glib::RefPtr< Gdk::Pixbuf >, const Gtk::TreePath );

			virtual float AlphaStrength( const Math::PolarPoint ) const;

			virtual void StartSculpt();
			virtual void ContinueSculpt();
		protected:
			virtual float factor_() const;

			template< typename T,
				  typename Action > void symmetrical_sculpt_( T& t, Action action )
			{
				Mesh& mesh( MeshHistory::Instance().GetCurrentMesh() );

				// Do action
				( t.*action )( mesh.GetMirror( false, false, false ) );
				if( Symmetry().X() )
					( t.*action )( mesh.GetMirror( true, false, false ) );
				if( Symmetry().Y() )
					( t.*action )( mesh.GetMirror( false, true, false ) );
				if( Symmetry().Z() )
					( t.*action )( mesh.GetMirror( false, false, true ) );
				if( Symmetry().X() && Symmetry().Y() )
					( t.*action )( mesh.GetMirror( true, true, false ) );
				if( Symmetry().Y() && Symmetry().Z() )
					( t.*action )( mesh.GetMirror( false, true, true ) );
				if( Symmetry().X() && Symmetry().Z() )
					( t.*action )( mesh.GetMirror( true, false, true ) );
				if( Symmetry().X() && Symmetry().Y() && Symmetry().Z() )
					( t.*action )( mesh.GetMirror( true, true, true ) );
			}

			virtual void pre_start_sculpt_();
			virtual void start_sculpt_( const EditData& );
			virtual void post_start_sculpt_();

			virtual void pre_continue_sculpt_();
			virtual void continue_sculpt_( const EditData& );
			virtual void post_continue_sculpt_();
		private:
			DirectionData direction_;
			bool colorize_;
			bool mask_;
			unsigned radius_;
			float strength_, intensity_;
			bool flip_;
			float pressure_;
			bool radius_pressure_, strength_pressure_, intensity_pressure_;
			Color pigment_;
			static SymmetryData symmetry_;
			bool dot_, layer_;
			Glib::RefPtr< Gdk::Pixbuf > shape_;
			Gtk::TreePath shape_path_;
		};
	}
}

#endif
